"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/keep.py $
$Id: keep.py 30557 2008-03-10 19:24:28Z rmasse $
"""
from durus.btree import BTree
from durus.persistent import PersistentObject
from durus.persistent_dict import PersistentDict
from qp.lib.spec import either, datetime_with_tz, mapping, spec, integer
from qp.lib.spec import require, get_spec_problems, anything, specify
from qp.lib.util import rand_int
from qp.lib.tz import UTC

class Keyed (object):
    """
    An item with an int key used as a key in an Keep.
    Note that the key attribute is set when the item is put in the Keep,
    so there is no set_key() method.
    This is a mixin for PersistentObject classes.
    """
    key_is = integer

    __slots__ = []

    def __init__(self):
        assert isinstance(self, PersistentObject)
        self.key = None

    def get_key(self):
        return self.key


class KeyCounter (PersistentObject):
    """
    The base class.  Generates keys using random 64 bit ints.
    """
    def next(self):
        """() -> integer
        This base implementation just uses a big random number.
        """
        return rand_int(64)


class Counter (KeyCounter):
    """
    Generates keys by incrementing them.
    """
    next_available_is = integer

    __slots__ = ['next_available']

    def __init__(self):
        self.next_available = 1

    def next(self):
        result = self.next_available
        self.next_available = self.next_available + 1
        return result


class GrowingRandomCounter (KeyCounter):
    """
    Keep a key counter that generates keys randomly, and gradually
    expands the numbers of bits in the random numbers.
    The idea is to take advantage of randomly distributed keys, but
    to try to keep the numbers relatively small.
    """
    bits_is = integer
    used_is = integer

    __slots__ = ['bits', 'used']

    def __init__(self):
        specify(self, bits=10, used=0)

    def next(self):
        self.used = self.used + 1
        # If more than 1% of the possible values have been used,
        # double the space of possible values.
        if (self.used / (2 ** self.bits)) > 0.01:
            self.bits += 1
        return rand_int(self.bits)


class Keep (object):
    """
    A simple database that stores a mapping of objects by key.
    This is a mixin for PersistentObject classes.
    """
    value_spec_is = spec(
        anything,
        "Specifies the type of values allowed in mapping")
    mapping_is = mapping({integer:Keyed}, either(PersistentDict, BTree))
    key_counter_is = KeyCounter

    __slots__ = []

    def __init__(self, value_spec=Keyed, mapping_class=BTree):
        assert isinstance(self, PersistentObject)
        self.mapping = mapping_class()
        self.value_spec = value_spec
        self.set_counter(Counter())

    def set_counter(self, counter):
        specify(self, key_counter=counter)

    def __len__(self):
        return len(self.mapping)

    def get_mapping(self):
        return self.mapping

    def get(self, key, default=None):
        return self.mapping.get(key, default)

    def itervalues(self):
        for value in self.mapping.itervalues():
            yield value

    def iterkeys(self):
        for key in self.mapping.iterkeys():
            yield key

    def iteritems(self):
        for item in self.mapping.iteritems():
            yield item

    def add(self, value):
        require(value, self.value_spec)
        assert value.key is None
        for attempt in range(1000):
            value.key = self.key_counter.next()
            if value.key not in self.mapping:
                break
        else:
            raise KeyError('Unable to find unused key for %r.' % value)
        if get_spec_problems(value):
            raise TypeError(''.join(get_spec_problems(value)))
        self.mapping[value.key] = value

    def remove(self, value):
        require(value, self.value_spec)
        assert value.key is self.mapping[value.key].key
        del self.mapping[value.key]


class Stamped (object):
    """
    Provides a timestamp.
    This is a mixin for PersistentObject classes.  
    """
    stamp_is = datetime_with_tz

    __slots__ = []

    def __init__(self):
        assert isinstance(self, PersistentObject)
        self.set_stamp()

    def get_stamp(self):
        return self.stamp

    def set_stamp(self):
        self.stamp = UTC.now()


def stamp_sorted(stamped_sequence):
    items = [(x.stamp, x) for x in stamped_sequence]
    return [x for (stamp, x) in sorted(items)]

def reverse_stamp_sorted(stamped_sequence):
    result = stamp_sorted(stamped_sequence)
    result.reverse()
    return result




#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/test/utest_util.py $
$Id: utest_util.py 30554 2008-03-10 17:43:53Z dbinger $
"""
from qp.lib.util import import_object, randbytes, rand_int, rand_str
from sancho.utest import UTest

class Test(UTest):

    def check_import_object(self):
        import sys
        assert sys is import_object('sys')
        assert UTest is import_object('sancho.utest.UTest')

    def check_rand(self):
        for x in range(1,100):
            assert 0 <= rand_int(x) < 2 ** x
            assert type(rand_str(x)) is str
            randbytes(x)

if __name__ == '__main__':
    Test()



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/hub/test/utest_web.py $
$Id: utest_web.py 30453 2008-01-11 11:48:05Z dbinger $
"""
from durus.utils import BytesIO, as_bytes
from qp.hub.web import ns_read_size, ns_reads, read_env
from sancho.utest import UTest, raises

class TestNetstring(UTest):

    def a(self):
        input = BytesIO()
        size = ns_read_size(input)
        assert size == None

    def b(self):
        input = BytesIO(as_bytes('4:0000,'))
        size = ns_read_size(input)
        assert size == 4, size

    def b2(self):
        input = BytesIO(as_bytes('42:0000,'))
        size = ns_read_size(input)
        assert size == 42, size

    def c(self):
        input = BytesIO(as_bytes('40000,'))
        raises(IOError, ns_read_size, input)

    def c(self):
        input = BytesIO(as_bytes('x40000:,'))
        raises(ValueError, ns_read_size, input)

    def d(self):
        input = BytesIO()
        result = ns_reads(input)
        assert result == None, result

    def e(self):
        input = BytesIO(as_bytes('4:0000,'))
        result = ns_reads(input)
        assert result == as_bytes('0000'), result

    def f(self):
        input = BytesIO(as_bytes('4:000,'))
        raises(IOError, ns_reads, input)

    def g(self):
        input = BytesIO(as_bytes('4:'))
        raises(IOError, ns_reads, input)

    def h(self):
        input = BytesIO(as_bytes('4:0000'))
        raises(IOError, ns_reads, input)

    def k(self):
        input = BytesIO(as_bytes('4:0000000'))
        raises(IOError, ns_reads, input)

    def l(self):
        input = BytesIO(as_bytes('4:0000,000'))
        result = ns_reads(input)
        assert result == as_bytes('0000'), result

    def m(self):
        input = BytesIO(as_bytes('8:a\0b\0c\0d\0,asdfasdf'))
        result = read_env(input)
        assert result == dict(a='b', c='d')

if __name__ == '__main__':
    TestNetstring()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/__init__.py $
$Id: __init__.py 30570 2008-03-11 14:44:15Z dbinger $
Copyright (c) Corporation for National Research Initiatives 2008. All Rights Reserved.
"""
__version__ = "2.1"


"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/lib/util.py $
$Id: util.py 31157 2008-09-18 18:22:13Z dbinger $
"""
from binascii import hexlify
from datetime import datetime
import os
import sys

if sys.version < "3":
    from cStringIO import StringIO
    from urlparse import urljoin
    from commands import getoutput
    from rfc822 import Message as message_from_file
else:
    from io import StringIO
    from urllib.parse import urljoin
    from subprocess import getoutput
    from email import message_from_file

if sys.version < "2.6":
    from sha import new as sha1
else:
    from hashlib import sha1

_used = [StringIO, sha1, urljoin, message_from_file] # to quiet code checker.

def as_str(x):
    if type(x) is type(''):
        return x
    else:
        return x.decode('latin1')

def import_object(name):
    """
    Import and return the object with the given name.
    """
    i = name.rfind('.')
    if i != -1:
        module_name = name[:i]
        object_name = name[i+1:]
        __import__(module_name)
        try:
            return getattr(sys.modules[module_name], object_name)
        except AttributeError:
            pass
    __import__(name)
    return sys.modules[name]


def randbytes(bytes):
    """Return bits of random data as a hex string."""
    return hexlify(os.urandom(bytes))

def rand_str(n):
    return as_str(randbytes(n))

def rand_int(bits):
    bytes, remainder = divmod(bits, 8)
    if remainder == 0:
        return int(randbytes(bytes), 16)
    else:
        return int(randbytes(bytes + 1), 16) >> (8 - remainder)

def trace(f):
    def trace(*args, **kwargs):
        start = datetime.now() # no tz, ok
        name = getattr(f, 'func_name', None) or getattr(f, '__name__')
        s = '%s(' % name
        for arg in args:
            s += "%r, " % arg
        for k, v in kwargs.items():
            s += "%s=%r, " % (k, v)
        if args or kwargs:
            s = s[:-2]
        s += ')'
        print("\n" + s)
        e = None
        try:
            result = f(*args, **kwargs)
        except:
            e = sys.exc_info()[0]
        duration = datetime.now() - start # no tz, ok
        print('%s finished in %s -> %r' % (s, duration, e or result))
        if e is not None:
            print(e)
            raise e
        return result
    return trace

def get_memory_usage():
    output = getoutput("ps -p %s -orsz,vsz" % os.getpid())
    rsz, vsz = output.split()[-2:]
    return int(rsz), int(vsz)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/hub/dispatcher.py $
$Id: dispatcher.py 31228 2008-10-23 01:46:46Z dbinger $

For systems that support passfd, this provides
a pre-forking server that uses file descriptor passing to off-load
requests to child worker processes.

"""
try:
    from qp.hub import passfd
    from qp.hub.passfd import recvfd, sendfd, socketpair
except ImportError:
    # If passfd is not installed, we can still run in non-forking mode.
    passfd = None

from durus.utils import as_bytes
from qp.lib.util import as_str
from select import select, error as select_error
from errno import EPIPE, EWOULDBLOCK, EINTR
from fcntl import fcntl, F_GETFL, F_SETFL
from socket import SOCK_STREAM, AF_UNIX, SOL_SOCKET, SO_REUSEADDR, AF_INET
from socket import socket, fromfd, MSG_PEEK, error as socket_error
from stat import ST_MTIME
from os import close, getpid, fork, O_NONBLOCK, read, write, waitpid, stat
from os import environ, WNOHANG, execve, kill
import re
import signal
import sys

def log(s):
    sys.stdout.write("hub[%s] %s\n" % (getpid(), s))
    sys.stdout.flush()

def select_ready(fds):
    try:
        return select(fds, [], [], None)[0]
    except select_error:
        e = sys.exc_info()[1]
        if e[0] == EINTR:  # got a signal, try again
            return []
        raise

class Child (object):
    """
    This represents a child process that is running a service loop.
    """
    __slots__ = ['pid', 'fd', 'client_address', 'ready_for_connection']

    def __init__(self, pid, child_fd):
        self.pid = pid
        self.fd = child_fd
        self.client_address = None
        self.ready_for_connection = False

    def fileno(self):
        return self.fd

    def close(self):
        """
        Close the connection to the child process.
        """
        try:
            close(self.fd)
        except OSError:
            log("close() failed for %s" % self.pid)

    def wait(self):
        """
        Wait for the child process to terminate.
        """
        log("Wait for %s" % self.pid)
        waitpid(self.pid, 0)
        log("%s is done." % self.pid)

    def kill(self):
        try:
            kill(self.pid, signal.SIGKILL)
        except OSError:
            log("kill() failed for %s" % self.pid)
        self.close()

    def is_ready_for_connection(self):
        return self.ready_for_connection

    def update_ready_for_connection(self):
        if not self.ready_for_connection:
            ready_byte = ''
            try:
                ready_byte = read(self.fd, 1)
            except socket_error:
                exc = sys.exc_info()[1]
                if exc[0] != EWOULDBLOCK:
                    raise
            except (OSError, IOError):
                pass
            self.ready_for_connection = (ready_byte == as_bytes('1'))

    def send_connection(self, conn, connection_address):
        if not self.is_ready_for_connection():
            return False
        try:
            sendfd(self.fd, conn.fileno())
        except IOError:
            exc = sys.exc_info()[1]
            if exc.errno == EPIPE:
                return False
            else:
                raise
        self.client_address = connection_address
        self.ready_for_connection = False
        conn.close()
        return True


class HubServer (object):

    banned_msg = (
        'HTTP/1.0 403 Forbidden\r\n'
        'Content-Type: text/plain\r\n\r\n'
        'Your IP address has been banned from accessing this site.\n'
        'Please email the webmaster here if you need the ban removed.')

    def __init__(self, site_name, max_children=5, banned=None,
        busy_limit=1):
        """(site_name:string,
            max_children:int=5,
            banned:str|None=None,
            busy_limit:int=1)
        max_children is the maximum number of child processes.  If the value
        of max_children is 0, the server runs in non-forking mode.
        banned is the path to a file, if any, containing IP addresses from
        which connections are not permitted.
        busy_limit is the maximum number of processes that may be kept
        busy by connections from any single IP address.
        """
        self.site_name = site_name
        if passfd:
            self.max_children = max_children
        else:
            # We can't pass file descriptors.
            # Fall back to non-forking mode.
            log("No passfd installed.  Using non-forking mode.")
            self.max_children = 0
        self.children = []
        self.banned = banned
        self.banned_time = -1
        self.banned_addresses = set()
        self.connection_queue = []
        self.busy_limit = busy_limit

    def spawn_child(self):
        """
        Unless we have reached the maximum number, start up a new child
        process.
        """
        if len(self.children) < self.max_children:
            parent_fd, child_fd = socketpair(AF_UNIX, SOCK_STREAM)
            # make child fd non-blocking
            flags = fcntl(child_fd, F_GETFL, 0)
            fcntl(child_fd, F_SETFL, flags | O_NONBLOCK)
            pid = fork()
            if pid == 0: # child
                close(child_fd)
                for s in self.listening_sockets + self.connection_queue:
                    s.close()
                import qp.hub.web
                args = [sys.executable, qp.hub.web.__file__,
                    self.site_name, str(parent_fd)]
                execve(sys.executable, args, environ)
            close(parent_fd)
            self.children.append(Child(pid, child_fd))
            log("Child [%s] %s of %s started." % (
                pid, len(self.children), self.max_children))

    def reap_children(self):
        """
        Check to see if any of the child processes have died, and if so,
        remove them from the list.
        """
        while self.children:
            (pid, status) = waitpid(-1, WNOHANG)
            if pid <= 0:
                break
            for child in self.children:
                if child.pid == pid:
                    log("reap %s" % pid)
                    close(child.fd)
                    self.children.remove(child)
                    break

    def do_stop(self):
        """
        This may be called from a signal handler to stop the child processes.
        """
        log("stopping children")
        for child in self.children:
            child.close()
        for child in self.children:
            child.wait()
        self.children = []
        log("children stopped")

    remote_address_re = re.compile('.*REMOTE_ADDR.([.\d]+)')
    x_forwarded_for_re = re.compile('.*X_FORWARDED_FOR: ([.\d]+)')

    def get_connection_address(self, connection):
        try:
            headers = connection.recv(2048, MSG_PEEK)
        except socket_error:
            log("in get_connection_address: %s" % sys.exc_info()[1])
            return None
        if not headers:
            return None
        headers = as_str(headers)
        match = (self.remote_address_re.match(headers) or
                 self.x_forwarded_for_re.match(headers))
        if match:
            return match.group(1)
        else:
            return connection.getpeername()[0]

    def delegated(self, connection):
        """
        Try to find a suitable child process to handle this connection,
        and pass it to the child.  If sucessful, return True.
        Otherwise, return False and wait for another chance.
        """
        connection_address = self.get_connection_address(connection)
        if connection_address in self.banned_addresses:
            # Shut up.  Go away.
            connection.send(self.banned_msg)
            connection.close()
            return True
        children_with_this_address = []
        if connection_address in (None, '127.0.0.1'):
            for child in self.children:
                if child.send_connection(connection, None):
                    return True
        else:
            children_with_this_address = [
                child for child in self.children
                if child.client_address == connection_address]
            if children_with_this_address:
                for child in children_with_this_address:
                    if child.send_connection(connection, connection_address):
                        return True
            if len(children_with_this_address) < self.busy_limit:
                for child in self.children:
                    if (child.client_address != connection_address and
                        child.send_connection(connection, connection_address)):
                        return True
        # Maybe a child process died?
        self.reap_children()
        if len(children_with_this_address) < self.busy_limit:
            # Spawn a new child if we haven't reached the max_children limit.
            self.spawn_child()
        return False

    def listen(self, *addresses):
        """
        Set self.listening_sockets to a list of sockets, bound the the given
        addresses and listening for new connections.
        To allow quick restarts without re-binding sockets, this method will,
        instead of doing the usual thing, use an environment variable to identify
        file descriptors for sockets that are presumed to already be bound to the
        given set of addresses.
        """
        name = 'QP_HUBSERVER_FILE_DESCRIPTORS'
        sockets = []
        addresses = [a for a in addresses if a]
        if environ.get(name):
            fds = [int(c) for c in environ.get(name).split(',')]
            assert len(fds) == len(addresses)
            sockets = [fromfd(fd, AF_INET, SOCK_STREAM)
                for fd in fds]
        if not sockets:
            def get_socket_bound_to_address(address):
                s = socket(AF_INET, SOCK_STREAM)
                s.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
                s.bind(address)
                s.listen(40)
                return s
            sockets = [get_socket_bound_to_address(address)
                for address in addresses]
        environ[name] = ','.join([str(s.fileno()) for s in sockets])
        self.listening_sockets = sockets

    def update_banned_addresses(self):
        """
        Consult the file named by self.banned, if any, and update
        self.banned_addresses to reflect the current content of the
        file.
        """
        if self.banned:
            try:
                mtime = stat(self.banned)[ST_MTIME]
            except OSError:
                pass
            else:
                if mtime > self.banned_time:
                    self.banned_time = mtime
                    f = open(self.banned)
                    new_banned_addresses = set(
                        [a.strip() for a in f.read().split()])
                    f.close()
                    for address in (self.banned_addresses -
                                    new_banned_addresses):
                        log('UNBANNED: %s' % address)
                    for address in (new_banned_addresses -
                                self.banned_addresses):
                        log('BANNED: %s' % address)
                    self.banned_addresses = new_banned_addresses

    def remove_hogs(self):
        connections_by_ip_address = {}
        for c in self.connection_queue:
            address = self.get_connection_address(c)
            if address not in connections_by_ip_address:
                connections_by_ip_address[address] = []
            connections_by_ip_address[address].append(c)
        def value_length(x):
            return len(x[1])
        by_frequency = sorted(
            connections_by_ip_address.items(), key=value_length)
        hog_address, hogs = by_frequency[-1]
        log("REMOVE %s HOGS FROM %s" % (len(hogs), hog_address))
        try:
            for c in hogs:
                c.close()
        except:
            pass # Don't let the server die for this.
        for child in list(self.children):
            if child.client_address == hog_address:
                child.kill()
                self.children.remove(child)
        self.connection_queue = [
            c for c in self.connection_queue if c not in hogs]

    def run(self, handle_connection=None):
        """
        The main service loop of the parent.
        """
        while True:
            if not handle_connection and not self.children:
                self.spawn_child()
            if len(self.connection_queue) > 50:
                self.remove_hogs()
            # Accept any new connections.
            to_check = self.children + self.listening_sockets
            self.update_banned_addresses()

            if len(self.connection_queue) > 0:
                log("connections in queue: %s." % len(self.connection_queue))
            # Wait until there is something to read.
            ready = select_ready(to_check)
            for s in ready:
                if s in self.listening_sockets:
                    try:
                        conn, addr = s.accept()
                    except socket_error:
                        e = sys.exc_info()[1]
                        if e[0] != EINTR: # signal
                            raise
                    self.connection_queue.append(conn)
            # Go ahead and read the ready-byte from children that
            # appear to have sent it.
            for child in self.children:
                if child in ready:
                    child.update_ready_for_connection()
            # Now try to actually handle connections in the queue.
            if handle_connection and self.max_children == 0:
                # We are in non-forking mode.  Handle waiting connections now.
                for c in self.connection_queue:
                    handle_connection(c)
                    c.close()
                self.connection_queue = []
            elif self.connection_queue:
                self.connection_queue = [
                    c for c in self.connection_queue
                    if not self.delegated(c)]


def worker(handle_connection, parent_fd):
    parent_fd = int(parent_fd)
    while True:
        try:
            # Tell the parent that we are ready.
            write(parent_fd, as_bytes("1"))
            # Receive the descriptor for the connection
            fd = recvfd(parent_fd)
        except (IOError, OSError):
            # The parent probably exited
            # (EPIPE comes thru as OSError).
            log("Terminating %s." % getpid())
            raise SystemExit
        # Make a blocking socket from the file descriptor.
        conn = fromfd(fd, AF_INET, SOCK_STREAM)
        conn.setblocking(1)
        close(fd)
        handle_connection(conn)
        conn.close()




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/fill/test/utest_form.py $
$Id: utest_form.py 30874 2008-06-25 03:38:51Z dbinger $
"""
from durus.logger import direct_output
from durus.utils import BytesIO, as_bytes
from qp.fill.form import Form, get_javascript_code
from qp.fill.widget import MultipleSelectWidget, CheckboxWidget, IntWidget
from qp.pub.common import clear_publisher, get_request
from qp.pub.hit import Hit
from qp.pub.publish import DurusPublisher
from sancho.utest import UTest, raises
import sys

class FormTest(UTest):

    def _pre(self):
        direct_output(sys.stdout)
        clear_publisher()

    def _post(self):
        clear_publisher()

    def a(self):
        publisher = DurusPublisher()
        hit = Hit(None, {})
        publisher.process_hit(hit)
        form = Form()
        form.add_submit('a')
        form.add_reset('b')
        form.add_hidden('c')
        form.add_string('d')
        form.add_text('e')
        form.add_checkbox('f')
        form.add_single_select('g', options=[0])
        form.add_multiple_select('h', options=[0])
        form.add_radiobuttons('i', options=[0])
        form.add_float('j')
        form.add_int('k')
        form.render()
        assert not form.is_submitted()

    def b(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        publisher.process_hit(hit)
        form = Form()
        form.add_submit('a')
        form.add_string('d')
        assert form.is_submitted()
        form.render()
        assert form.get('d') == '1'
        form.clear()
        assert form.get('d') is None

    def c(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        publisher.process_hit(hit)
        form = Form(method="get")
        form.add_submit('a')
        form.add_string('d')
        assert form.is_submitted()
        form.render()

    def d(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="POST"))
        publisher.process_hit(hit)
        form = Form(id='ok')
        form.add_submit('a')
        form.add_string('d')
        form.set_error('a', 'err')
        assert not form.is_submitted()
        assert 'id="ok"' in form.render()

    def e(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        hit.get_request().process_inputs()
        publisher.set_hit(hit)
        form = Form(method="get")
        form.add_submit('a')
        form.add_string('d')
        form.has_key('d')
        assert form.get('d') == '1'
        assert form['d'] == '1'
        assert form.is_submitted()
        get_javascript_code()['a'] = 'javascript'
        assert 'javascript' in str(form.render())

    def f(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        hit.get_request().process_inputs()
        publisher.set_hit(hit)
        form = Form(method="get")
        form.add(MultipleSelectWidget, 'a', options=[range(10)])
        class FakePersistent(object): pass
        a = FakePersistent()
        b = FakePersistent()
        a._p_oid = '11111111'
        b._p_oid = '22222222'
        FakePersistent.__str__ = lambda x: 'ok'
        FakePersistent._p_format_oid = lambda x: x._p_oid
        form.add(MultipleSelectWidget, 'b', options = [a, b])
        raises(AssertionError, # because descriptions are not unique.
               form.add, MultipleSelectWidget, 'c', options = [a, b, 1])

    def g(self):
        publisher = DurusPublisher()
        hit = Hit(None, dict(REQUEST_METHOD="GET", QUERY_STRING='d=1'))
        hit.get_request().process_inputs()
        publisher.set_hit(hit)
        form = Form(method="get")
        form.add(MultipleSelectWidget, 'a', options=[range(10)])
        form.add_string('d')
        assert not form.get_widget('a').is_submitted()
        assert form.get_widget('d').is_submitted()
        assert form.is_submitted()
        assert form.get_widget('a').is_submitted()

class WidgetTest (UTest):

    def _pre(self):
        self.publisher = DurusPublisher()

    def _post(self):
        self.publisher = None
        clear_publisher()

    def _form(self, input='', **env):
        input = as_bytes(input)
        env.setdefault('CONTENT_LENGTH', len(input))
        env.setdefault('CONTENT_TYPE', 'application/x-www-form-urlencoded')
        hit = Hit(BytesIO(input), env)
        hit.get_request().process_inputs()
        self.publisher.set_hit(hit)
        return Form(method=env.get('REQUEST_METHOD').lower())

class CheckboxWidgetTest (WidgetTest):

    def a(self):
        form = self._form(REQUEST_METHOD="GET", QUERY_STRING='d=yes')
        form.add(CheckboxWidget, 'd')
        assert form.get_widget('d').is_submitted()
        assert form['d'] == True
        form = self._form(input='d=yes', REQUEST_METHOD="POST")
        form.add(CheckboxWidget, 'd')
        assert form.get_widget('d').is_submitted()
        assert form['d'] == True

    def b(self):
        form = self._form(REQUEST_METHOD="GET", QUERY_STRING='d=1')
        form.add(CheckboxWidget, 'd')
        assert form.get_widget('d').is_submitted()
        assert form['d'] == False, form['d']
        form = self._form(input='d=1', REQUEST_METHOD="POST")
        form.add(CheckboxWidget, 'd')
        assert form.get_widget('d').is_submitted()
        assert form['d'] == False, form['d']

class IntWidgetTest (WidgetTest):

    def a(self):
        form = self._form(REQUEST_METHOD="GET", QUERY_STRING='d=24')
        form.add(IntWidget, 'd')
        assert form.get_widget('d').is_submitted()
        assert form['d'] == 24
        form = self._form(input='d=24', REQUEST_METHOD="POST")
        form.add(IntWidget, 'd')
        assert form.get_widget('d').is_submitted(), get_request().get_fields()
        assert form['d'] == 24, get_request().get_fields()

    def b(self):
        form = self._form(REQUEST_METHOD="GET", QUERY_STRING='d=')
        form.add(IntWidget, 'd')
        assert form.get_widget('d').is_submitted()
        assert form['d'] == None, form['d']
        form = self._form(input='d=', REQUEST_METHOD="POST")
        form.add(IntWidget, 'd')
        assert form.get_widget('d').is_submitted()
        assert form['d'] == None, form['d']

if __name__ == '__main__':
    FormTest()
    CheckboxWidgetTest()
    IntWidgetTest()



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/__main__.py $
$Id: __main__.py 30561 2008-03-10 22:20:04Z dbinger $
"""
from qp.lib.site import Site
import sys
from os import kill
from signal import SIGHUP
from optparse import OptionParser
import qp.sites

parser = OptionParser()
parser.set_usage("qp [options] [start|stop|restart] [site]*")
parser.set_description('''\
Control the servers for the sites in your qp installation.
No options is the same as "--status".
"stop" is the same as "--stop".
"start" is the same as "--start".
"restart" is the same as "--stop --start".

If "--stop" and "--start" are both present and the durus server
is already running, only the web server is restarted.

The script searches for sites in the qp.sites package, which
in turn searches through the following list, by default.
    ''' + "\n    ".join(path for path in qp.sites.sites_path) + "\n")

def format_description(self, *args):
    return self.parser.get_description()
parser.format_description = format_description

parser.add_option(
    '-u', '--start', dest='start', default=False, action='store_true',
    help="Start the durus and web servers.")
parser.add_option(
    '-d', '--stop', dest='stop', default=False, action='store_true',
    help="Stop the durus and web servers.")
parser.add_option(
    '-q', '--quickrestart', dest='quickrestart', action='store_true',
    help='Quick restart of the web server(s) of the named sites.')
parser.add_option(
    '-v', '--status', dest='status', default=False, action='store_true',
    help='Show the current status of the durus and web servers.')
parser.add_option(
    '-c', '--configuration', dest='show', default=False, action='store_true',
    help='Show the site configuration.')
parser.add_option(
    '-l', '--log', dest='site_log', action='store_true',
    help='tail -f the log for the last named site.')
parser.add_option(
    '-i', '--interact', dest='site_interact', action='store_true',
    help='Open an interactive session with the last named site.')

(options, args) = parser.parse_args()

if 'start' in args:
    options.stop = True
    options.start = True
    args.remove('start')
if 'stop' in args:
    options.stop = True
    args.remove('stop')
if 'restart' in args:
    options.start = True
    options.stop = True
    args.remove('restart')

sites = []

if args:
    for site_name in args :
        site = Site(site_name)
        try:
            site.get_configuration()
            sites.append(site)
        except ImportError:
            e = sys.exc_info()[1]
            print("\nImport failed for site %s (%s).\n" % (site_name, e))
else:
    # This imports all sites.
    sites = list(Site.get_sites().values())

if not sites:
    print("Could not find sites in any of these places:")
    for path in qp.sites.__path__:
        print("   " + path)
    raise SystemExit

for site in sites:
    if options.stop:
        site.stop_web()
        site.stop_durus()
    if options.start:
        site.start_durus()
        site.start_web()
    if options.quickrestart:
        web_pid = site.is_web_running()
        if web_pid:
            kill(web_pid, SIGHUP)
    if options.show:
        site.show()
    if (options.status or not (
        options.stop or
        options.start or
        options.site_log or
        options.site_interact or
        options.quickrestart or
        options.show)):
        site.status()

if options.site_interact:
    sites[-1].interaction()

if options.site_log:
    sites[-1].log_tail()



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qp/http/test/utest_response.py $
$Id: utest_response.py 31364 2008-11-26 01:58:22Z dbinger $
"""
from durus.utils import BytesIO, as_bytes, join_bytes
from qp.http.response import HTTPResponse, Stream
from qp.lib.spec import unicode_string
from sancho.utest import UTest
import sys

class Test(UTest):

    def a(self):
        response = HTTPResponse()
        response.set_content_type('text/html', 'utf-8')
        response.set_status(200)
        response.set_status(200, 'Excellent')
        assert response.get_status_code() == 200
        assert response.get_status() == (200, 'Excellent')
        response.set_status(479)
        try:
            response.set_status(600)
            assert 0
        except ValueError:
            pass
        try:
            response.set_status('200')
            assert 0
        except TypeError:
            pass
        response.set_expires(None)
        response.set_expires(seconds=200)
        response.set_header('foo', 1)
        assert response.get_header('foo') == 1
        response.set_cookie('a', 23)
        response.expire_cookie('a')
        response.set_body('ok')
        response.set_compress(False)
        response.set_body('ok' * 1000)

    def b(self):
        response = HTTPResponse()
        assert response.get_mime_type() == 'text/html'
        response.set_content_type('text/plain', 'utf-8')
        assert response.get_mime_type() == 'text/plain'
        assert (200, 'OK') == response.status
        assert response.get_content_length() == None
        response.set_body('ab')
        assert response.get_content_length() == 2
        response.set_body(Stream(BytesIO(as_bytes('a')), 1))
        assert response.get_content_length() == 1
        response.set_cookie('cookie', 1)
        headers = response.generate_headers()
        headers_dict = dict(headers)
        del headers_dict[as_bytes('Date')]
        assert len(headers_dict) == 4
        for k, v in [('Content-Type', 'text/plain; charset=utf-8'),
                     ('Expires', '-1'),
                     ('Content-Length', '1'),
                     ('Set-Cookie', 'cookie="1"')]:
            assert headers_dict[as_bytes(k)] == as_bytes(v)            
            
        response.set_expires(None)
        headers = response.generate_headers()
        response.set_expires(days=1)
        headers = response.generate_headers()
        response.set_cookie('a', 'x"', secure=1, expires=1,
                            b=None, a='all"ok"')
        response.set_header('bogus', 'yes')
        headers = response.generate_headers()

    def d(self):
        response = HTTPResponse()
        response.set_cookie('a', 'x')
        response.set_body('ok')
        headers = response.generate_headers()
        headers_dict = dict(headers)
        del headers_dict[as_bytes('Date')]
        assert len(headers_dict) == 4
        for k, v in [('Content-Type', 'text/html; charset=utf-8'),
                     ('Expires', '-1'),
                     ('Content-Length', '2'),
                     ('Set-Cookie', 'a="x"')]:
            assert headers_dict[as_bytes(k)] == as_bytes(v)

    def check_write(self):
        stream = Stream(BytesIO(as_bytes('a')), 1)
        assert join_bytes(stream) == as_bytes('a')
        response = HTTPResponse()
        response.buffered=False
        out = BytesIO()
        response.write(out, include_status=False)
        response.write(out, include_status=True)
        response.write(out, include_body=False)
        response.set_body('a')
        response.write(out)
        response.set_body(stream)
        response.write(out)
        response.set_body(Stream(BytesIO(as_bytes('a'*1000)), 1000))
        response.write(out)
        response.set_body(unicode_string('0') * 1000)
        response.write(out)

    def c(self):
        response = HTTPResponse()
        response.set_cookie('a', '"ok', path="/", secure=True)
        response.set_cookie('b', None)
        response.set_header('ok', 'this')
        list(response.generate_headers())
        assert response.cache == 0
        response.set_expires(1)
        list(response.generate_headers())
        response.cache = None
        list(response.generate_headers())

if __name__ == '__main__':
    Test()
